/**
 * Sort lines in ascending/descending order
 *
 * Can sort lines numrical or string based. Sort at a specific column/position on lines.
 * Plugin is callable in 2 ways:
 * By using the GUI or by hotkey, using last used settings saved by the GUI.
 *
 * @category  WeBuilder Plugin
 * @package   Sort Lines
 * @author    Peter Klein <pmk@io.dk>
 * @copyright 2016 Peter Klein
 * @license   http://www.freebsd.org/copyright/license.html  BSD License
 * @version   1.0
 */

/**
 * [CLASS/FUNCTION INDEX of SCRIPT]
 *
 *     55   function QuickSort(SL, left, right, startPos, endPos, sortOrder, ignoreCase, numericSort)
 *     86   function Swap(SL, i, j)
 *    103   function GetCompareValue(line, startPos, endPos)
 *    122   function Compare(a, b, startPos, endPos, ignoreCase, numericSort)
 *    146   function ShowSortOptions(Sender)
 *    330   function TEdit1OnExit(Sender)
 *    346   function TEdit2OnExit(Sender)
 *    356   function SortLines(Sender)
 *    410   function OnInstalled()
 *
 * TOTAL FUNCTIONS: 9
 * (This index is automatically created/updated by the WeBuilder plugin "DocBlock Comments")
 *
 */

/**
 * Location of settings ini file
 *
 * @var settingsFile
 */
var settingsFile = Script.GetPath + "settings.ini";

/**
 * Quicksort
 *
 * @param  object    SL: TStringList
 * @param  integer   left: start index of TStringList
 * @param  integer   right: end index of TStringList
 * @param  integer   startPos: Start position of text to compare
 * @param  integer   endPos: End position of text to compare (-1 = EOL)
 * @param  integer   sortOrder: sorting order. -1 = Ascending, 1 = Descending
 * @param  boolean   ignoreCase: case insensetive sorting
 * @param  boolean   numericSort: use numeric sorting
 *
 * @return object    TStringList
 */
function QuickSort(SL, left, right, startPos, endPos, sortOrder, ignoreCase, numericSort) {

    if (left < right) {
        var pivot = right,
            pivotValue = SL[pivot],
            partitionIndex = left;

        for (var i = left; i < right; i++) {
            if (Compare(SL[i],pivotValue, startPos, endPos, ignoreCase, numericSort) == sortOrder ) {
                SL = Swap(SL, i, partitionIndex);
                partitionIndex++;
            }
        }
        SL = Swap(SL, right, partitionIndex);

        //sort left and right
        SL = QuickSort(SL, left, partitionIndex - 1, startPos, endPos, sortOrder, ignoreCase, numericSort);
        SL = QuickSort(SL, partitionIndex + 1, right, startPos, endPos, sortOrder, ignoreCase, numericSort);
    }
    return SL;
}

/**
 * Swap TStringList lines
 *
 * @param  object   TStringList
 * @param  integer   i
 * @param  integer   j
 *
 * @return object   TStringList
 */
function Swap(SL, i, j) {
   var t = SL[i];
   SL[i] = SL[j];
   SL[j] = t;
   return SL;
}

/**
 * Get part of line to do comparison on
 *
 * @param  string    line
 * @param  integer   startPos
 * @param  integer   endPos
 * @param  boolean   ignoreCase
 *
 * @return string
 */
function GetCompareValue(line, startPos, endPos) {
    var len = Length(line);
    if (startPos > len) startPos = len;
    if ((endPos > len) || (endPos == -1)) endPos = len;
    return Copy(line, startPos, endPos);
}

/**
 * Compare 2 strings
 *
 * @param  string   a
 * @param  string   b
 * @param  integer  startPos
 * @param  integer  endPos
 * @param  boolean  ignoreCase
 * @param  boolean  numericSort
 *
 * @return mixed  LessThanValue = -1, EqualsValue =	0, GreaterThanValue = 1
 */
function Compare(a, b, startPos, endPos, ignoreCase, numericSort) {
    a = GetCompareValue(a, startPos, endPos);
    b = GetCompareValue(b, startPos, endPos);
    if ((!numericSort) && (ignoreCase)) {
        return CompareText(a, b);
    }
    else if (numericSort) {
        if (ValidFloat(a)) a = StrToFloat(a);
        else a = 0;
        if (ValidFloat(b)) b = StrToFloat(b);
        else b = 0;
    }
    if (a < b) return -1;
    else if (a > b) return 1;
    else return 0;
}

/**
 * Show GUI
 *
 * @param  object   Sender
 *
 * @return void
 */
function ShowSortOptions(Sender) {
    var settings = new TIniFile(settingsFile);
    var sortOrder = settings.ReadInteger("Settings","SortOrder", 1);
    var ignoreCase = settings.ReadBool("Settings", "IgnoreCase", false);
    var numericSort = settings.ReadBool("Settings", "NumericSort", false);
    var startPos = settings.ReadInteger("Settings", "StartPos", 0);
    var endPos = settings.ReadInteger("Settings", "EndPos", -1);

    var mleft = 14,
        mtop = 14,
        vspace = 32,
        hdiv = 10,
        lspace = 4,
        bspace = 16,
        topPos = 0,
        leftPos = 0,
        gTopPos = 0,
        gLeftPos = 0;
        margin = 6;

    var modalForm = new TForm(WeBuilder);
    modalForm.Caption = "Advanced Sort Options";
    modalForm.ClientHeight = 300;
    modalForm.ClientWidth = 250;
    modalForm.Color = clBtnFace;
    modalForm.Font.Color = clWindowText;
    modalForm.Font.Height = -14;
    modalForm.Font.Name = "Tahoma";
    modalForm.Position = poScreenCenter;
    modalForm.BorderIcons = biSystemMenu; //remove maximize & minimize buttons
    modalForm.BorderStyle = bsSingle; //disable dialog resizing
    modalForm.PixelsPerInch = 120;

    var groupWidth = modalForm.ClientWidth - margin - margin - 8;
    gTopPos = margin;

    GroupBox1 = new TGroupBox(WeBuilder);
    GroupBox1.Parent = modalForm;
    GroupBox1.Caption = "Sort Order";
    GroupBox1.Anchors = akLeft + akTop;
    GroupBox1.Width = groupWidth;
    GroupBox1.Left = margin;
    GroupBox1.Top = gTopPos;

    topPos = mtop + mtop;
    leftPos = mleft;

    RadioButton1 = new TRadioButton(WeBuilder);
    RadioButton1.Parent = GroupBox1;
    RadioButton1.Caption = "Ascending";
    RadioButton1.Checked = (sortOrder == 1);
    RadioButton1.SetBounds(leftPos, topPos, 90, 17);

    leftPos += RadioButton1.Width + lspace;

    RadioButton2 = new TRadioButton(WeBuilder);
    RadioButton2.Parent = GroupBox1;
    RadioButton2.Caption = "Descending";
    RadioButton2.Checked = (sortOrder == -1);
    RadioButton2.SetBounds(leftPos, topPos, 95, 17);

    GroupBox1.Height = topPos + mtop + mtop + 13;

    gTopPos += GroupBox1.Height + margin;

    GroupBox2 = new TGroupBox(WeBuilder);
    GroupBox2.Parent = modalForm;
    GroupBox2.Caption = "Options";
    GroupBox2.Anchors = akLeft + akTop;
    GroupBox2.Width = groupWidth;
    GroupBox2.Left = margin;
    GroupBox2.Top = gTopPos;

    topPos = mtop + mtop;
    leftPos = mleft;

    CheckBox1 = new TCheckBox(WeBuilder);
    CheckBox1.Parent = GroupBox2;
    CheckBox1.Caption = "Ignore case";
    CheckBox1.Checked = ignoreCase;
    CheckBox1.SetBounds(leftPos, topPos, 97, 17);

    leftPos += CheckBox1.Width + bspace;

    CheckBox2 = new TCheckBox(WeBuilder);
    CheckBox2.Parent = GroupBox2;
    CheckBox2.Caption = "Numeric sort";
    CheckBox2.Checked = numericSort;
    CheckBox2.SetBounds(leftPos, topPos, 97, 17);

    GroupBox2.Height = topPos + mtop + mtop + 13;

    gTopPos += GroupBox2.Height + margin;

    GroupBox3 = new TGroupBox(WeBuilder);
    GroupBox3.Parent = modalForm;
    GroupBox3.Caption = "Sort Position";
    GroupBox3.Width = groupWidth;
    GroupBox3.Left = margin;
    GroupBox3.Top = gTopPos;

    topPos = mtop + mtop;
    leftPos = mleft;

    Label1 = new TLabel(WeBuilder);
    Label1.Parent = GroupBox3;
    Label1.Alignment = taRightJustify;
    Label1.Caption = "Start Column:";
    Label1.SetBounds(leftPos, topPos + 3, 90, 17);

    leftPos += Label1.Width + lspace;

    Edit1 = new TEdit(WeBuilder);
    Edit1.Parent = GroupBox3;
    Edit1.Text = startPos;
    Edit1.OnExit = "TEdit1OnExit";
    Edit1.SetBounds( leftPos, topPos, 60, 25);

    topPos += vspace;
    leftPos = mleft;

    Label2 = new TLabel(WeBuilder);
    Label2.Parent = GroupBox3;
    Label2.Alignment = taRightJustify;
    Label2.Caption = "End Column:";
    Label2.ShowHint = true;
    Label2.Hint = "Column end position. -1 equals end of line";
    Label2.SetBounds(leftPos, topPos + 3, 90, 17);

    leftPos += Label2.Width + lspace;

    Edit2 = new TEdit(WeBuilder);
    Edit2.Parent = GroupBox3;
    Edit2.Text = endPos;
    Edit2.OnExit = "TEdit2OnExit";
    Edit2.ShowHint = true;
    Edit2.Hint = "Column end position. -1 equals end of line";
    Edit2.SetBounds( leftPos, topPos, 60, 25);

    GroupBox3.Height = topPos + mtop + mtop + 13;

    gTopPos += GroupBox3.Height + margin;

    CancelButton = new TButton(WeBuilder);
    CancelButton.Parent = modalForm;
    CancelButton.Caption = "Cancel";
    CancelButton.ModalResult = mrCancel;
    CancelButton.SetBounds(modalForm.Width - margin - mleft - 75 - 75 - bspace, gTopPos, 75, 33);

    OKButton = new TButton(WeBuilder);
    OKButton.Parent = modalForm;
    OKButton.Caption = "Sort";
    OKButton.ModalResult = mrOk;
    OKButton.SetBounds(modalForm.Width - margin - mleft - 75 , gTopPos, 75, 33);

	var modalResult = modalForm.ShowModal;
    if (modalResult == mrOK) {
        sortOrder = 1;
        if (RadioButton2.Checked) sortOrder = -1;
        settings.WriteInteger("Settings", "SortOrder", sortOrder);
        settings.WriteBool("Settings", "IgnoreCase", CheckBox1.Checked);
        settings.WriteBool("Settings", "NumericSort", CheckBox2.Checked);

        TEdit1OnExit(Edit1); // make sure value is valid integer
        settings.WriteInteger("Settings", "StartPos", StrToInt(Edit1.Text));
        TEdit2OnExit(Edit2); // make sure value is valid integer
        settings.WriteInteger("Settings", "EndPos", StrToInt(Edit2.Text));

        SortLines(Sender);
    }

    // Remove TIniFile object
    delete settings;
    // Remove Modal object
    delete modalForm;
}

/**
 * OnExit event handler
 *
 * @param  object   Sender
 *
 * @return void
 */
function TEdit1OnExit(Sender) {
    if (!ValidInt(Sender.Text)) {
        Sender.Text = "0";
    }
    else if (StrToInt(Sender.Text) < 0) {
        Sender.Text = "0";
    }
}

/**
 * OnExit event handler
 *
 * @param  object   Sender
 *
 * @return void
 */
function TEdit2OnExit(Sender) {
    if (!ValidInt(Sender.Text)) {
        Sender.Text = "-1";
    }
    else if (StrToInt(Sender.Text) < -1) {
        Sender.Text = "-1";
    }
}


function SortLines(Sender) {
    var uSel = Editor.Selection;
    if (Trim(Editor.SelText) == "") {
        // No selection found
        if (Script.ReadSetting("Sort all", "0") == "1") {
            // Set selection to editor content
            uSel.SelStartLine = 0;
            uSel.SelEndLine = Editor.LineCount - 1;
        }
        else {
            return;
        }
    }

    // Test if in Block Select Mode
    var blockMode = Actions.IsChecked("ActBlockSelect");
    if (blockMode) {
        // Turn Block Select Mode off while doing the insert
        Actions.Execute("ActBlockSelect");
    }

    // Expand selection to encapsulate complete lines
    uSel.SelStartCol = 0;
    uSel.SelEndCol = Length(Editor.Lines[uSel.SelEndLine]);
    Editor.Selection = uSel;

    // Get settings
    var settings = new TIniFile(settingsFile);
    var sortOrder = settings.ReadInteger("Settings","SortOrder", 1);
    var ignoreCase = settings.ReadBool("Settings", "IgnoreCase", false);
    var numericSort = settings.ReadBool("Settings", "NumericSort", false);
    var startPos = settings.ReadInteger("Settings", "StartPos", 0);
    var endPos = settings.ReadInteger("Settings", "EndPos", -1);
    delete settings;

    var SL = new TStringList;
    SL.Text = Editor.SelText;
    SL = QuickSort(SL, 0, SL.Count-1, startPos, endPos, sortOrder, ignoreCase, numericSort);

    Editor.SelText = SL.Text;

    delete SL;

    if (blockMode) {
        // Turn Block Select Mode back on if it was active before insert
        Actions.Execute("ActBlockSelect");
    }
}

/**
 * Show info when plugin is installed
 *
 * @return void
 */
function OnInstalled() {
  alert("Sort Lines 1.0 by Peter Klein installed sucessfully!");
}

Script.ConnectSignal("installed", "OnInstalled");
var bmp = new TBitmap;
var icon = LoadFileToBitmap(Script.Path + "sort.png", bmp);
var act = Script.RegisterDocumentAction("", "Sort Lines", "Shift+Ctrl+O", "SortLines");
Actions.SetIcon(act, bmp);
icon = LoadFileToBitmap(Script.Path + "sort_edit.png", bmp);
act = Script.RegisterDocumentAction("", "Advanced Sort Options", "", "ShowSortOptions");
Actions.SetIcon(act, bmp);
delete bmp;